/*
 * tcpstream - Firestorm TCP state tracking implementation.
 * Copyright (c) 2002 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL v2
 *
 * TODO:
 *  o Fix wait states
 *  o Implement timeouts for SYN states
 *  o Stream reassembly
 *  o Support sack
 *  o Connection re-sync
 *  o Connection pickup ?
*/

#define TCP_PAWS_24DAYS (60 * 60 * 24 * 24)
#define TCP_PAWS_MSL 60
#define TCP_PAWS_WINDOW 60

#if 1
#define dmesg(x...)
#define __INLINE__ static inline
#else
#define dmesg mesg
#define __INLINE__
static char *state_str[]={
	"CLOSED",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT1",
	"FIN_WAIT2",
	"TCP_TIME_WAIT",
	"TCP_CLOSE",
	"TCP_CLOSE_WAIT",
	"TCP_LAST_ACK",
	"TCP_LISTEN",
	"TCP_CLOSING"
};
#endif

#include <cleanup.h>
#include "tcpip.h"

/* Configuration */
unsigned int tcp_stateful=0;
unsigned int tcp_minttl=0;
unsigned int tcp_numstreams=1024;
unsigned int max_concurrent=0;

/* Alert generator */
struct generator tcpstream_gen=init_generator("tcpstream", NULL);

/* Configuration arguments */
struct arg tcpstream_args[]={
	{"minttl", ARGTYPE_PUINT, NULL, {vp_uint:&tcp_minttl}},
	{"num_streams", ARGTYPE_PBYTES, NULL, {vp_bytes:&tcp_numstreams}},
	{NULL, ARGTYPE_NOP, NULL}
};

/* TCP segment variables */
struct tcpseg {
	u_int32_t ack,seq,win,len,seq_end;
	u_int32_t tsval;
	int saw_tstamp;
};

/* Main tcpstream data structures:
 * Hash table for lookups -- O(n) in pathological cases, can be O(1)
 * LRU list to evict oldest items -- O(1)
 * Stack for allocation -- O(1)
 */
unsigned int tcp_hashsz=512;
struct tcp_session *all_sessions=NULL;
struct tcp_session **tcp_hash=NULL;

struct tcp_session *tcp_latest=NULL;
struct tcp_session *tcp_oldest=NULL;

struct tcp_session **alloc_stack=NULL;
unsigned int next_session=0;

/* Statistics counters */
unsigned int tcp_packets=0;
unsigned int tcp_state_errs=0;
unsigned int tcp_broadcast=0;
unsigned int tcp_lowttl=0;
unsigned int tcp_timeouts=0;

/* Wrap-proof seq/ack calculations */
__INLINE__ int before(u_int32_t s1, u_int32_t s2) {
	return (int32_t)(s1-s2) < 0;
}
__INLINE__ int after(u_int32_t s1, u_int32_t s2) {
	return (int32_t)(s2-s1) < 0;
}
__INLINE__ int between(u_int32_t s1, u_int32_t s2, u_int32_t s3) {
	return s3 - s2 >= s1 - s2; /* is s2<=s1<=s3 ? */
}
__INLINE__ u_int32_t tcp_receive_window(struct tcp_stream *tp)
{
	int32_t win=tp->rcv_wup+tp->rcv_wnd-tp->rcv_nxt;
	if ( win<0 )
		win=0;
	return (u_int32_t)win;
}
__INLINE__ int tcp_sequence(struct tcp_stream *tp, u_int32_t seq, u_int32_t end_seq)
{
	return !before(end_seq, tp->rcv_wup) &&
		!after(seq, tp->rcv_nxt+tcp_receive_window(tp));
}

/* Just for timestamps */
int tcp_fast_options(struct pkt_tcphdr *t, u_int32_t *tsval)
{
	char *tmp, *end;

	/* Return if we don't have any */
	if ( t->doff<<2 <= sizeof(struct pkt_tcphdr)) return 0;

	/* Work out where they begin and end */
	tmp=end=(char *)t;
	tmp+=sizeof(struct pkt_tcphdr);
	end+=(t->doff<<2);

	while ( tmp<end ) {
		if ( *tmp == TCPOPT_EOL || *tmp == TCPOPT_NOP ) {
			tmp++;
			continue;
		}

		if ( !(tmp+1 < end) ) break;

		switch(*tmp) {
		case TCPOPT_TIMESTAMP:
			if ( !(tmp+10 < end) ) break;
			*tsval=ntohl(*((u_int32_t *)(tmp+2)));
			return 1;
		}

		tmp+=*(tmp+1);
	}

	return 0;
}

/* This will parse TCP options for SYN packets */
void tcp_syn_options(struct tcp_stream *s, struct pkt_tcphdr *t, u_int32_t sec)
{
	char *tmp, *end;

	/* Return if we don't have any */
	if ( t->doff<<2 <= sizeof(struct pkt_tcphdr)) return;

	/* Work out where they begin and end */
	tmp=end=(char *)t;
	tmp+=sizeof(struct pkt_tcphdr);
	end+=(t->doff<<2);

	while ( tmp<end ) {
		if ( *tmp == TCPOPT_EOL || *tmp == TCPOPT_NOP ) {
			tmp++;
			continue;
		}

		if ( !(tmp+1 < end) ) break;

		switch(*tmp) {
		case TCPOPT_SACK_PERMITTED:
			s->flags|=TF_SACK_OK;
			break;
		case TCPOPT_TIMESTAMP:
			s->flags|=TF_TSTAMP_OK;

			/* Only check the bit we want */
			if ( !(tmp+10 < end) ) break;

			s->ts_recent=ntohl(*((u_int32_t *)(tmp+2)));
			s->ts_recent_stamp=sec;

			break;
		case TCPOPT_WSCALE:
			if ( !(tmp+2 < end) ) break;
			s->flags|=TF_WSCALE_OK;

			/* rfc1323: must log error and limit to 14 */
			if ( (s->scale=*(tmp+2)) > 14 ) s->scale=14;
			break;
		}

		tmp+=*(tmp+1);
	}
}

/* Perform a state transition */
__INLINE__ void transition(struct tcp_session *t, int c, int s)
{
	char cip[32],sip[32];
	struct in_addr ca, sa;

	t->client.state=c;
	t->server.state=s;

	ca.s_addr=t->c_addr;
	sa.s_addr=t->s_addr;
	strncpy(cip, inet_ntoa(ca), sizeof(cip));
	strncpy(sip, inet_ntoa(sa), sizeof(sip));

	dmesg(M_DEBUG, "tcpstream: %s:%u(%s) -> %s:%u(%s)",
		cip, ntohs(t->c_port), state_str[c],
		sip, ntohs(t->s_port), state_str[s]);
}

/* Hash function. Hashes to the same bucket even
 * when source and destinations are inverted */
__INLINE__ int tcp_hashfn(u_int32_t saddr, u_int32_t daddr,
	u_int16_t sport, u_int16_t dport)
{
	int h = ((saddr ^ sport) ^ (daddr ^ dport));
	h ^= h>>16;
	h ^= h>>8;
	return h % tcp_hashsz;
}

/* Push a tcp session back on top of the stack */
void tcp_free(struct tcp_session *s)
{
	if ( !s || !next_session ) {
		mesg(M_DEBUG,"tcpstream: severe insanity");
		return;
	}

	if ( s->next )
		s->next->pprev=s->pprev;
	*s->pprev=s->next;

	if ( s->next_time )
		s->next_time->prev_time=s->prev_time;
	if ( s->prev_time )
		s->prev_time->next_time=s->next_time;

	if ( s==tcp_oldest ) tcp_oldest=s->prev_time;
	if ( s==tcp_latest ) tcp_latest=s->next_time;

	next_session--;
	alloc_stack[next_session]=s;
}


__INLINE__ struct proto *tcp_find_proto(u_int16_t port)
{
	struct proto_child *pc;

	/* IDs are server ports */
	for(pc=tcp_p.children; pc; pc=pc->next)
	{
		if ( port == pc->id ) return pc->proto;
	}

	return NULL;
}

/* Add a new session, never fails */
struct tcp_session *tcp_new(struct pkt_iphdr *iph,
	struct pkt_tcphdr *tcph, struct packet *p)
{
	struct tcp_session *ret;

	/* If we have run out of connections, then
	 * just free the oldest one in the LRU */
	if ( next_session >= tcp_numstreams ) {
		tcp_free(tcp_oldest);
		tcp_timeouts++;
	}

	/* Pop next free item off of the stack */
	ret=alloc_stack[next_session++];

	/* Keep track of max number of concurrent sessions
	 * seen so far */
	if ( next_session > max_concurrent )
	       max_concurrent=next_session;

	/* Setup the key values */
	ret->c_addr=iph->saddr;
	ret->s_addr=iph->daddr;
	ret->c_port=tcph->sport;
	ret->s_port=tcph->dport;
	ret->bucket=tcp_hashfn(ret->c_addr, ret->s_addr,
		ret->c_port, ret->s_port);

	/* Setup initial state */
	transition(ret, TCP_SYN_SENT, 0);
	ret->server.isn=ntohl(tcph->seq);
	ret->client.snd_una=ret->server.isn+1;
	ret->client.snd_nxt=ret->client.snd_una+1;
	ret->client.rcv_wnd=ntohs(tcph->win);

	ret->server.rcv_nxt=ret->client.snd_una;
	ret->server.rcv_wup=ret->client.snd_una;

	/* server sees clients initial options */
	tcp_syn_options(&ret->server, tcph, p->time.tv_sec);

	/* Link in to LRU list */
	ret->next_time=tcp_latest;
	if ( tcp_latest ) tcp_latest->prev_time=ret;
	tcp_latest=ret;
	ret->prev_time=NULL;
	if ( !tcp_oldest ) tcp_oldest=ret;

	/* Link in to the hashtable */
	if ( (ret->next=tcp_hash[ret->bucket]) ) {
		ret->next->pprev=&ret->next;
	}
	tcp_hash[ret->bucket]=ret;
	ret->pprev=&tcp_hash[ret->bucket];

	ret->related=NULL;
	ret->proto=tcp_find_proto(ntohs(ret->s_port));

	return ret;
}

/* Find a TCP session given a packet */
struct tcp_session *tcp_find(struct pkt_iphdr *iph,
	struct pkt_tcphdr *tcph,
	int *to_server)
{
	struct tcp_session *s;
	int idx=tcp_hashfn(iph->saddr, iph->daddr,
		tcph->sport, tcph->dport);

	for(s=tcp_hash[idx]; s; s=s->next) {
		if (	s->s_addr==iph->saddr &&
			s->c_addr==iph->daddr &&
			s->s_port==tcph->sport &&
			s->c_port==tcph->dport ) {
			*to_server=0;
			return s;
		}
		if (	s->c_addr==iph->saddr &&
			s->s_addr==iph->daddr &&
			s->c_port==tcph->sport &&
			s->s_port==tcph->dport ) {
			*to_server=1;
			return s;
		}
	}

	return NULL;
}

__INLINE__ void tcp_established(struct tcp_stream *snd,
	struct tcp_stream *rcv,
	u_int32_t seq,
	u_int32_t ack,
	u_int32_t win)
{
	if ( after(ack, rcv->snd_una) && !before(ack, rcv->snd_nxt) ) {
		rcv->snd_una=ack;
		rcv->rcv_nxt=seq+1;
		snd->rcv_wup=ack;
		snd->rcv_wnd=win;
	}

	snd->snd_una=seq;
	snd->snd_nxt=seq+1;
}

/* A segment contained valid data so attempt
 * stream reassembly -- this is not finished yet */
void tcpstream_data(
	struct tcp_session *s,
	struct tcp_stream *rcv,
	struct pkt_tcphdr *tcph,
	struct tcpseg *tp)
{
	if ( tp->seq==rcv->rcv_nxt ) {
		/* TODO: How do we deal with this? */
		if (tcp_receive_window(rcv) == 0)
			return;

		/* TODO: copy in to buffer */
		dmesg(M_DEBUG, "In order: %u->%u",
			tp->seq-rcv->isn,
			tp->seq_end-rcv->isn);

		rcv->rcv_nxt=tp->seq_end;
	}else{
		/* A retransmit */
		if ( !after(tp->seq_end,rcv->rcv_nxt) )
			return;

		dmesg(M_DEBUG, "Out of order: %u->%u",
			tp->seq-rcv->isn,
			tp->seq_end-rcv->isn);
	}
}

/* Process incoming TCP packets here */
__INLINE__ int tcpstream_tcpseg(struct packet *p, unsigned int i)
{
	struct pkt_iphdr *iph=p->layer[i].h.ip;
	struct pkt_tcphdr *tcph=p->layer[i+1].h.tcp;
	struct tcp_session *s;
	struct tcp_stream *snd, *rcv;
	int to_server;
	int ret=0;
	struct tcpseg tp;

	/* Host order header values */
	tp.ack=ntohl(tcph->ack);
	tp.seq=ntohl(tcph->seq);
	tp.win=ntohs(tcph->win);
	tp.len=ntohs(iph->tot_len) - (iph->ihl<<2) - (tcph->doff<<2);
	tp.seq_end=tp.seq+tp.len+tcph->flags.bits.syn+tcph->flags.bits.fin;
	tp.tsval=0;
	tp.saw_tstamp=0;

	/* Find the session this packet comes from */
	if ( !(s=tcp_find(iph, tcph, &to_server)) ) {
		if ( (tcph->flags.bits.syn) &&
			!(tcph->flags.bits.ack) &&
			!(tcph->flags.bits.rst)) {

			/* SYN: part 1 of connection handshake */
			p->layer[i+1].session=tcp_new(iph, tcph, p);
			p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_2SVR;
			return ret;
		}

		dmesg(M_DEBUG,"tcpstream: never heard of you");
		return ret;
	}else p->layer[i+1].session=s;

	/* figure out who is sender and who is reciever */
	if ( to_server ) {
		snd=&s->client;
		rcv=&s->server;
		p->layer[i+1].flags|=FLAG_TCP_2SVR;
	}else{
		snd=&s->server;
		rcv=&s->client;
		p->layer[i+1].flags&=(~FLAG_TCP_2SVR);
	}

	/* Move to head of LRU list */
	if ( s->next_time ) s->next_time->prev_time=s->prev_time;
	if ( s->prev_time ) s->prev_time->next_time=s->next_time;
	if ( s == tcp_oldest ) tcp_oldest=s->prev_time;
	if ( s == tcp_latest ) tcp_latest=s->next_time;
	s->next_time=tcp_latest;
	s->prev_time=NULL;
	if ( !tcp_oldest ) tcp_oldest=s;
	if ( tcp_latest ) tcp_latest->prev_time=s;
	tcp_latest=s;

	/* MTF of hash list */
	if ( s->next )
		s->next->pprev=s->pprev;
	*s->pprev=s->next;
	if ( (s->next=tcp_hash[s->bucket]) )
		s->next->pprev=&s->next;
	tcp_hash[s->bucket]=s;
	s->pprev=&tcp_hash[s->bucket];

	/* Deal with a SYN/ACK */
	if ( rcv->state==TCP_SYN_SENT ) {
		/* fist check the ack field */
		if ( tcph->flags.bits.ack ) {
		 	/* if SND.UNA =< SEG.ACK =< SND.NXT
			 * then ACK is acceptable */
			if ( !(between(tp.ack, rcv->snd_una, rcv->snd_nxt)) ) {
				return ret;
			}

			p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;
		}

		/* then check the rst flag */
		if ( tcph->flags.bits.rst ) {
			transition(s, 0, 0);
			return 1;
		}

		if ( tcph->flags.bits.syn ) {
			/* update the advertised window */
			snd->rcv_wnd=tp.win;

			/* update sequencing information */
			snd->snd_una=tp.seq+1;
			snd->snd_nxt=snd->snd_una+1;

			rcv->rcv_nxt=tp.seq+1;
			rcv->rcv_wup=tp.seq+1;
			rcv->isn=tp.seq;
			if ( tcph->flags.bits.ack )
				rcv->snd_una=tp.ack;

			/* client now sees servers initial options */
			tcp_syn_options(&s->client, tcph, p->time.tv_sec);

			/* Check whether to use window scaling */
			if (!(rcv->flags&TF_WSCALE_OK && snd->flags&TF_WSCALE_OK)) {
				rcv->scale=0;
				snd->scale=0;
			}

			/* Success */
			transition(s, TCP_SYN_SENT, TCP_SYN_RECV);
		}

		return ret;
	}

	/* First check the sequence number */
	if ( !tcp_sequence(rcv, tp.seq, tp.seq_end) ) {
		dmesg(M_DEBUG,"tcpstream: got out of sequence segment");
		return ret;
	}

	/* we now know that this packet is in-state */
	if ( rcv->state ) {
		p->layer[i+1].flags|=FLAG_TCP_STATE|FLAG_TCP_SURE;
	}else{
		p->layer[i+1].flags|=FLAG_TCP_STATE;
	}

	/* rfc1323: H1. Apply PAWS checks first */
	if ( rcv->flags&TF_TSTAMP_OK &&
		(tp.saw_tstamp=tcp_fast_options(tcph, &tp.tsval)) ) {
		/* TODO: Fix PAWS check, its buggy as fuck */
		if ( (int32_t)(rcv->ts_recent - tp.tsval) > TCP_PAWS_WINDOW &&
			p->time.tv_sec<rcv->ts_recent_stamp + TCP_PAWS_24DAYS ) {
			mesg(M_DEBUG, "PAWS discard: %u->%u %u->%u",
				rcv->ts_recent, tp.tsval,
				p->time.tv_sec, rcv->ts_recent_stamp + TCP_PAWS_24DAYS);
			return ret;
		}
	}

	/* Second, check the RST bit */
	if ( tcph->flags.bits.rst ) {
		dmesg(M_DEBUG,"TCP stream was reset");
		transition(s, 0, 0);
		return 1;
	}

	/* rfc1323: PAWS: update ts_recent */
	if ( tp.saw_tstamp && !after(tp.seq,rcv->rcv_wup) ) {
		if((int32_t)(tp.tsval - rcv->ts_recent) >= 0 ||
		   p->time.tv_sec >= rcv->ts_recent_stamp + TCP_PAWS_24DAYS) {
			rcv->ts_recent=tp.tsval;
			rcv->ts_recent_stamp=p->time.tv_sec;
		}
	}

	/* Third, check security and precendece (pfft) */

	/* Fourth, check the SYN bit */
	if ( tcph->flags.bits.syn && !before(tp.seq,rcv->rcv_nxt) ) {
		dmesg(M_DEBUG,"tcpstream: in window syn (wtf?!)");
		transition(s, 0, 0);
		return 1;
	}

	/* Scale the window */
	tp.win<<=snd->scale;

	/* Fifth, check the ack field */
	if ( !tcph->flags.bits.ack ) goto no_ack;
	switch(rcv->state) {
	case TCP_SYN_RECV:
		if ( rcv->snd_una <= tp.ack && tp.ack <= rcv->snd_nxt ) {
			transition(s, TCP_ESTABLISHED, TCP_ESTABLISHED);
			p->layer[i+1].flags|=FLAG_TCP_CT_EST;
			rcv->snd_una=tp.ack;
			snd->rcv_wnd=tp.win;
			snd->rcv_wup=tp.seq+1;
		}
		break;
	case TCP_ESTABLISHED:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		break;
	case TCP_FIN_WAIT1:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		rcv->state=TCP_FIN_WAIT2;
		snd->state=TCP_LAST_ACK;
		transition(s, s->client.state, s->server.state);
	case TCP_FIN_WAIT2:
		break;
	case TCP_CLOSE_WAIT:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		break;
	case TCP_CLOSING:
		tcp_established(snd,rcv,tp.seq,tp.ack,tp.win);
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_LAST_ACK:
		rcv->state=0;
		transition(s, s->client.state, s->server.state);
		ret=2;
		break;
	default:
		break;
	}

no_ack:
	/* sixth check URG bit */
	if ( !tcph->flags.bits.urg ) goto no_urg;

no_urg:
	/* seventh process the segment text */
	if ( !tp.len || !s->proto ) goto no_data;
	if ( (to_server && (s->proto->sflags&PSF_REQUEST)) ) goto reassemble;
	if ( (!to_server && (s->proto->sflags&PSF_REPLY)) ) goto reassemble;
	goto no_data;
reassemble:
	switch(rcv->state) {
	case TCP_ESTABLISHED:
	case TCP_FIN_WAIT1:
	case TCP_FIN_WAIT2:
		tcpstream_data(s, rcv, tcph, &tp);
	}


no_data:
	/* eighth, check the FIN bit */
	if ( !tcph->flags.bits.fin ) goto no_fin;
	rcv->rcv_nxt++;
	switch(rcv->state) {
	case TCP_SYN_RECV:
	case TCP_ESTABLISHED:
		rcv->state=TCP_CLOSE_WAIT;
		snd->state=TCP_FIN_WAIT1;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_FIN_WAIT1:
		/* if fin has been acked do time_wait else closing */
		rcv->state=TCP_TIME_WAIT;
		transition(s, s->client.state, s->server.state);
		break;
	case TCP_FIN_WAIT2:
		rcv->state=TCP_TIME_WAIT;
		snd->state=0;
		transition(s, s->client.state, s->server.state);
		ret=2;
		break;
	case TCP_CLOSE_WAIT:
	case TCP_CLOSING:
	case TCP_LAST_ACK:
	case TCP_TIME_WAIT:
		/* Remain in the same state */
		break;
	default:
		break;
	}

no_fin:

	if ( ret==2 ) {
		/* FIXME: this is fucked up because the user will
		 * see this packet with no state info, which is
		 * wrong, besides, we want a WAIT state so we can
		 * see packets transmitted within the 2MSL wait time
		 * as actually having state... */
		if ( rcv->state==TCP_TIME_WAIT ) rcv->state=0;
		if ( snd->state==TCP_TIME_WAIT ) snd->state=0;
		ret=(snd->state||rcv->state) ? 0 : 1;
	}

	return ret;
}

void icmperr_process(struct packet *p, unsigned int l)
{
	struct pkt_iphdr *outer_iph;
	struct pkt_icmphdr *icmph;
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	struct tcp_session *s;
	struct tcp_stream *hlf;
	int to_server;

	/* Get all the headers we need */
	if ( l+2 >= p->llen ) return;
	outer_iph=p->layer[l-1].h.ip;
	icmph=p->layer[l].h.icmp;
	iph=p->layer[l+1].h.ip;
	tcph=p->layer[l+2].h.tcp;

	/* Make sure the packet is cool */
	if ( iph->protocol!=6 ) return;
	if ( icmph->type != ICMP_DEST_UNREACH ) return;
	if ( icmph->code > NR_ICMP_UNREACH ) return;

	/* Check the IP packet came from the endpoint
	 * ( if necessary ) */
	if ( icmph->code==ICMP_PROT_UNREACH ||
		icmph->code==ICMP_PORT_UNREACH ) {
		if ( iph->saddr!=outer_iph->daddr ) {
			/* XXX: Alert here ? */
			return;
		}
	}

	/* Lookup our session */
	if ( !(s=tcp_find(iph, tcph, &to_server)) ) {
		/* XXX: Alert here ? */
		return;
	}

	/* Find out which side sent the offending packet */
	if ( iph->daddr==s->c_addr ) {
		hlf=&s->server;
	}else{
		hlf=&s->client;
	}

	/* Only reset opening connections */
	if ( hlf->state!=TCP_SYN_SENT && hlf->state!=TCP_SYN_RECV ) {
		/* XXX: Alert here ? */
		return;
	}

	/* TODO: Check seq against ISN */
	mesg(M_DEBUG,"TCP reset by ICMP: %u %u", htonl(hlf->isn), tcph->seq);
	transition(s, 0, 0);
	tcp_free(s);
}

void tcpstream_process(struct packet *pkt, unsigned int i)
{
	struct pkt_iphdr *iph=pkt->layer[i].h.ip;
	struct layer *l=&pkt->layer[i+1];
	struct tcp_session *s;
	int ret=0;

	l->flags|=FLAG_TCP_TRACK;

	/* Ignore fragments */
	if ( iph->frag_off & ipfmask )
		goto nodecode;

	/* Ignore bad IP checksums - ooh naughty! */
	if ( !(pkt->layer[i].flags&FLAG_IP_CSUM) )
		goto nodecode;

	/* Ignore bad TCP checksums - bad boy! */
	if ( !(l->flags&FLAG_TCP_CSUM) )
		goto nodecode;

	/* Ignore broadcast and multicast packets */
	if ( (pkt->flags & (FP_MULTICAST|FP_BROADCAST)) ) {
		tcp_broadcast++;
		goto nodecode;
	}

	/* Ignore packets with TTLs lower than minttl */
	if ( iph->ttl < tcp_minttl ) {
		tcp_lowttl++;
		goto nodecode;
	}

	/* Check we aren't in a icmp error message */
	if ( i>0 && pkt->layer[i-1].proto==&icmp_p )
		goto nodecode;

	/* Actually do the state tracking */
	ret=tcpstream_tcpseg(pkt, i);

	/* Statistics counters */
	if ( !(l->flags & FLAG_TCP_STATE) )
		tcp_state_errs++;
	tcp_packets++;

	/* Have we actually been decoded? */
	if ( !(s=l->session) ) goto nodecode;
	if ( !s->proto ) goto nodecode;

	/* Dispatch to the protocol handler */
	pkt->layer[pkt->llen].flags=0;
	pkt->layer[pkt->llen].session=NULL;
	pkt->layer[pkt->llen].proto=s->proto;
	s->proto->decode(pkt);
	goto done;

nodecode:
	if ( pkt->layer[pkt->llen].h.raw<pkt->end )
		pkt->layer[pkt->llen++].proto=NULL;
	dispatch(pkt);
done:
	if (ret) tcp_free(l->session);
	return;
}

void tcpstream_free(void)
{
	mesg(M_INFO,"tcpstream: max_concurrent=%u num_active=%u",
		max_concurrent, next_session);
	mesg(M_INFO,"tcpstream: %u state errors out of %u packets",
	       tcp_state_errs, tcp_packets);
	mesg(M_INFO,"tcpstream: %u broadcasts, %u ttl evasions, %u timeouts",
		tcp_broadcast, tcp_lowttl, tcp_timeouts);

	if ( tcp_hash ) free(tcp_hash);
	if ( alloc_stack ) free(alloc_stack);
	if ( all_sessions ) free(all_sessions);
}

int tcpstream_init(char *args)
{
	int i;

	if ( tcp_stateful ) {
		mesg(M_ERR,"tcpstream: Can't add tcpstream twice!");
		return 0;
	}

	if ( args && args_parse(tcpstream_args, args, NULL)<=0 ) {
		mesg(M_ERR,"tcpstream: parse error: %s", args);
		return 0;
	}

	if ( tcp_minttl > 255 ) {
		mesg(M_ERR,"tcpstream: minttl must be < 256");
		return 0;
	}

	if ( tcp_numstreams < 32 ) {
		mesg(M_WARN,"tcpstream: num_streams is stupidly "
			"low, setting to 512");
		tcp_numstreams=512;
	}

	/* Select a size for the hash table such that
	 * num_streams/hash_size is no larger than 128
	 * and such that the minimum size is 512 */
	for(tcp_hashsz=512;
		tcp_numstreams/tcp_hashsz>128;
		tcp_hashsz<<=1);

	/* Make sure its always odd */
	tcp_hashsz|=1;

	/* Initialise data structures */
	if ( !(all_sessions=calloc(tcp_numstreams,
		sizeof(struct tcp_session))) ) {
		goto oom;
	}
	if ( !(alloc_stack=calloc(tcp_numstreams,
		sizeof(struct tcp_session *))) ) {
		free(all_sessions);
		goto oom;
	}
	if ( !(tcp_hash=calloc(tcp_hashsz,
		sizeof(*tcp_hash))) ) {
		free(all_sessions);
		free(alloc_stack);
		goto oom;
	}
	next_session=0;
	memset(tcp_hash, 0, sizeof(tcp_hash));
	for(i=0; i<tcp_numstreams; i++) {
		alloc_stack[i]=&all_sessions[i];
	}

	mesg(M_INFO,"tcpstream: %u streams hashed into %u buckets (%u KB)",
		tcp_numstreams, tcp_hashsz,
		(tcp_numstreams*sizeof(struct tcp_session))/1024);
	tcp_stateful=1;

	return 1;
oom:
	mesg(M_ERR,"tcpstream: calloc(): %s", get_err());
	return 0;
}
